import asyncio
import json

import click

from higgsfield.internal.util import wd_path, parse_origin_link_or_else
from higgsfield.internal.cfg import AppConfig
from .setup import Setup

from base64 import b64decode


@click.command("get-hosts")
def hosts():
    wd = wd_path()
    app_config = AppConfig.from_path(wd)
    click.echo(",".join(app_config.hosts))


@click.command("get-nproc-per-node")
def proc_per_node():
    wd = wd_path()
    app_config = AppConfig.from_path(wd)
    click.echo(str(app_config.number_of_processes_per_node))


@click.command("get-ssh-details")
def ssh_details():
    wd = wd_path()
    app_config = AppConfig.from_path(wd)
    print(
        json.dumps(
            {
                "key": app_config.key,
                "user": app_config.user,
                "port": app_config.port,
                "hosts": ",".join(app_config.hosts),
            },
            indent=2,
        )
    )


@click.command("decode-secrets")
@click.argument("env", type=str, required=True)
def decode_secrets(env: str):
    env_path = wd_path() / "env"
    if env_path.exists():
        raise ValueError("env file already exists")

    env_path.write_text(b64decode(env.encode()).decode())




@click.command("setup-nodes")
def setup_nodes():
    wd = wd_path()
    app_config = AppConfig.from_path(wd)

    project_path = wd

    origin_url = app_config.get_git_origin_url(project_path)

    if origin_url is None:
        raise ValueError("Have you pushed your project to github?")

    origin_url = parse_origin_link_or_else(origin_url)

    if origin_url is None:
        raise ValueError("Please use ssh or https url for github repo.")

    app_config.github_repo_url = origin_url

    app_config.set_git_origin_url(project_path)

    setup = Setup(app_config, project_path)

    try:
        setup.create_ssh_key_file()
        asyncio.run(setup.setup_nodes())
    finally:
        setup.finish()



import asyncio
from pathlib import Path
from typing import List

from jinja2 import Environment, FileSystemLoader
from higgsfield.internal.cfg import AppConfig
import asyncssh

from higgsfield.internal.util import templates_path
from higgsfield.internal.experiment.builder import header


docker_install_script = '''/bin/bash -c "$(curl -fsSL https://gist.githubusercontent.com/arpanetus/1c1210b9e432a04dcfb494725a407a70/raw/5d47baa19b7100261a2368a43ace610528e0dfa2/install.sh)"'''
invoker_install_script = """wget https://github.com/ml-doom/invoker/releases/download/latest/invoker-latest-linux-amd64.tar.gz && \
        tar -xvf invoker-latest-linux-amd64.tar.gz && \
        sudo mv invoker /usr/bin/invoker && \
        rm invoker-latest-linux-amd64.tar.gz"""


def deploy_key_script(key: str, project_name: str, deploy_key_string: str):
    return f"""sudo mkdir -p ~/.ssh && \
    echo "{key}" > ~/.ssh/{project_name}-github-deploy.key && \
    chmod 600 ~/.ssh/{project_name}-github-deploy.key && \
    sudo touch ~/.ssh/config && \
    sudo chmod 644 ~/.ssh/config && \
    echo "{deploy_key_string}" | sudo tee -a ~/.ssh/config
    """


class Setup:
    app_config: AppConfig
    path: str
    deploy_key: str
    project_path: Path

    def __init__(
        self,
        app_config: AppConfig,
        project_path: Path,
    ):
        self.app_config = app_config

        if reason := self.app_config.is_valid() is not None:
            raise ValueError(reason)

        self.project_path = project_path

    def create_ssh_key_file(self):
        if self.app_config.key is None:
            raise ValueError("SSH_KEY in env is None")

        with Path.home() / ".ssh" / f"temp-{self.app_config.name}.key" as f:
            f.write_text(self.app_config.key)
            f.chmod(0o600)
            self.path = str(Path.resolve(f.absolute()))

    def finish(self):
        Path(self.path).unlink()

    async def establish_connections(self):
        if self.app_config.key is None:
            raise ValueError("SSH_KEY in env is None")

        self.connections: List[asyncssh.SSHClientConnection] = []
        for host in self.app_config.hosts:
            self.connections.append(
                await asyncssh.connect(
                    host,
                    port=self.app_config.port,
                    username=self.app_config.user,
                    client_keys=[self.path],
                )
            )

    def set_deploy_key(self):
        with Path.home() / ".ssh" / "higgsfield" / f"{self.app_config.name}-github-deploy.key" as f:
            self.deploy_key = f.read_text()

    def _build_deploy_key_string(self):
        return f"Host github.com-{self.app_config.name}\n\tHostName github.com\n\tIdentityFile ~/.ssh/{self.app_config.name}-github-deploy.key\n\tIdentitiesOnly yes\n\tStrictHostKeyChecking no\n\tUserKnownHostsFile=/dev/null\n\tLogLevel=ERROR\n"

    async def setup_nodes(self):
        await self.establish_connections()

        if len(self.connections) == 0:
            print("\n\n\nNO CONNECTIONS!!!\n\n\n")
        
        print("\n\n\nINSTALLING DOCKER\n\n\n")
        async def printer(thing):
            print(thing)
        to_run = []
        for conn in self.connections:
            to_run.append(printer(await conn.run(docker_install_script)))

        await asyncio.gather(*to_run)

        print("\n\n\nINSTALLING INVOKER\n\n\n")
        to_run = []
        for conn in self.connections:
            to_run.append(printer(await conn.run(invoker_install_script)))

        await asyncio.gather(*to_run)

        print("\n\n\nSETTING UP DEPLOY KEY\n\n\n")
        self.set_deploy_key()
        dk_script = deploy_key_script(
            self.deploy_key, self.app_config.name, self._build_deploy_key_string()
        )
        to_run = []
        for conn in self.connections:
            to_run.append(printer(await conn.run(dk_script)))
    
        await asyncio.gather(*to_run)
        
        print("\n\n\nPULLING BASE DOCKER IMAGE\n\n\n")
        to_run = []
        for conn in self.connections:
            to_run.append(printer(await conn.run(f"docker pull higgsfield/pytorch:latest")))

        await asyncio.gather(*to_run)


        print("\n\n\nSeems like everything is done by now. Go run your experiments.\n\n\n")




import torch
from torch.distributed.fsdp.sharded_grad_scaler import ShardedGradScaler

class Scaler(object):
    def __init__(self, model):
        if model.fsdp:
            self.scaler = ShardedGradScaler()
        else:
            return torch.cuda.amp.GradScaler()
 


import torch.distributed as dist

from torch.utils.data import (
    DistributedSampler, 
    DataLoader
)

from transformers import (
    LlamaTokenizer, 
    default_data_collator
)

from higgsfield.dataset import TorchCompletionDataset

class HiggsfieldSampler(DistributedSampler):
    def __init__(
        self, 
        dataset, 
        shuffle=True,
        seed=0, 
        drop_last=False
    ):
        rank=dist.get_rank()
        num_replicas=dist.get_world_size()
        
        super(HiggsfieldSampler, self).__init__(
            dataset=dataset, 
            num_replicas=num_replicas,
            rank=rank, 
            shuffle=shuffle,
            seed=seed, 
            drop_last=drop_last,
        )

class LlamaLoader(DataLoader):
    def __init__(
        self,
        dataset, 
        tokenizer=LlamaTokenizer.from_pretrained("meta-llama/Llama-2-7b-hf"),
        max_sequence_length=2048,
        batch_size_per_gpu=1,
        shuffle=True, 
        seed=0,
        num_workers=0, 
        pin_memory=False, 
        drop_last=False,
        timeout=0, 
        worker_init_fn=None,
        multiprocessing_context=None, 
        *, 
        prefetch_factor=None,
        persistent_workers=False,
        pin_memory_device=""
    ):
    
        dataset = TorchCompletionDataset(
            dataset,
            tokenizer,
            max_sequence_length,
        )
        
        sampler = HiggsfieldSampler(dataset, shuffle=shuffle, seed=seed,)
        
        super(LlamaLoader, self).__init__(
            dataset, 
            batch_size=batch_size_per_gpu,
            sampler=sampler,
            num_workers=num_workers,
            pin_memory=pin_memory, 
            drop_last=drop_last,
            timeout=timeout, 
            worker_init_fn=worker_init_fn,
            multiprocessing_context=multiprocessing_context, 
            prefetch_factor=prefetch_factor,
            persistent_workers=persistent_workers,
            pin_memory_device=pin_memory_device 
        )
        


import os
import shutil
from pathlib import Path

import time
import json
import torch
import torch.distributed as dist

from torch.distributed.fsdp.fully_sharded_data_parallel import (
    FullyShardedDataParallel as FSDP,
)

from optimum.bettertransformer import BetterTransformer

from .fsdp_utils import fsdp_model_state_dict_rank0


DEFAULT_CHECKPOINT_PATH = Path.home()  / \
                          ".cache" / \
                          "higgsfield" / \
                          os.environ["PROJECT_NAME"] / \
                          "experiments" / \
                          os.environ["EXPERIMENT_NAME"] / \
                          os.environ["RUN_NAME"]

class Checkpoint:
    '''
        Saving checkpoint to:
            ~/.cache/higgsfield/{project_name}/experiments/{experiment_name}/{run_name}
    '''
    def __init__(
        self,
        model,
        optimizer=None,
        lr_scheduler=None,
        scaler=None,
    ):
        '''
            model: Higgsfield.model
        '''
        if os.environ["PROJECT_NAME"] and os.environ["EXPERIMENT_NAME"] and os.environ["RUN_NAME"]:
            save_dir = DEFAULT_CHECKPOINT_PATH 
        else:
            raise NotImplementedError("Support single GPU/process not implemeted yet")
            

        self.save_dir     = save_dir
        self.model        = model
        self.optimizer    = optimizer
        self.lr_scheduler = lr_scheduler
        self.scaler       = scaler
    
    def save(self, epoch, steps=0, metadata={}):
        
        save_path = Path(self.save_dir) / f"epoch_{epoch}_steps_{steps}"
        save_path.mkdir(exist_ok=True, parents=True)
        
        model_path     = save_path / "model.pt"
        optimizer_path = save_path / "optimizer.pt"
        
        t0 = time.perf_counter()
        save_distributed_model_rank0(model_path, self.model)
        
        if self.optimizer:
            save_distributed_optimizer_rank0(optimizer_path, self.model, self.optimizer)
        t1 = time.perf_counter()

        if int(os.environ["LOCAL_RANK"]) == 0:
            print(f"State checkpoint of {steps} steps saved to {save_path}")
            print(f"Checkpoint Time = {t1-t0:.4f}\n")
            
            if self.lr_scheduler:
                lr_scheduler_path = save_path / "lr_scheduler.pt"
                torch.save(self.lr_scheduler.state_dict(), lr_scheduler_path)
            
            if self.scaler:
                scaler_path = save_path / "scaler.pt"
                torch.save(self.grad_scaler.state_dict(), scaler_path)
            
            metadata_path = save_path / "metadata.json"
            metadata["epoch"] = epoch
            metadata["steps"] = steps

            with open(metadata_path, "w+") as jsonFile:
                json.dump(metadata, jsonFile) 

def save_distributed_model_rank0(checkpoint_path, model):
    '''
        model: FSDP
    '''
    rank = dist.get_rank()
    
    cpu_state = fsdp_model_state_dict_rank0(model)
        
    if rank == 0:
        torch.save(cpu_state, checkpoint_path)
    
def save_distributed_optimizer_rank0(checkpoint_path, model, optimizer):
    '''
        model: FSDP
        optimizer: torch.optim
    '''
    rank = dist.get_rank()
    
    optim_state = FSDP.full_optim_state_dict(model, optimizer)

    if rank == 0:
        torch.save(optim_state, checkpoint_path)


from torch.distributed.fsdp.fully_sharded_data_parallel import (
    FullyShardedDataParallel as FSDP,
)

from torch.distributed.fsdp.api import (
    FullStateDictConfig,
    StateDictType,
)

fullstate_save_policy = FullStateDictConfig(
    offload_to_cpu=True,
    rank0_only=True,
)

def fsdp_model_state_dict_rank0(model):
    with FSDP.state_dict_type(
        model, StateDictType.FULL_STATE_DICT, fullstate_save_policy
    ):
        cpu_state = model.state_dict()
        
    return cpu_state


import torch
import gc

def get_tensors_from_gc(gpu_only=True):
    for obj in gc.get_objects():
        try:
            if torch.is_tensor(obj):
                tensor = obj
            elif hasattr(obj, "data") and torch.is_tensor(obj.data):
                tensor = obj.data
            else:
                continue
            
            if tensor.is_cuda or not gpu_only:
                yield tensor
        except Exception:  # nosec B112 pylint: disable=broad-exception-caught
            continue

def empty_cache():
    cnt = 0
    for obj in get_tensors_from_gc():
        obj.detach()
        obj.grad = None
        obj.untyped_storage().resize_(0)
    cnt += 1
    gc.collect()
    torch.cuda.empty_cache()


import click

from higgsfield.internal.cli import (
    init_cmd,
    ci,
    run_experiment,
    show_deploy_key,
    build_experiments,
    ci_cli,
)


@click.group()
def cli():
    """Higgsfield CLI"""
    pass


cli.add_command(init_cmd)
cli.add_command(ci)
cli.add_command(run_experiment)
cli.add_command(build_experiments)
cli.add_command(show_deploy_key)
cli.add_command(ci_cli.setup_nodes)



from typing import Any, Dict, List
from pathlib import Path

from fabric.config import os
from higgsfield.internal.experiment.decorator import _experiments
from higgsfield.internal.experiment.params import parse_kwargs_to_params
from higgsfield.internal.experiment.builder import _source_experiments


class Launch:
    experiment_name: str
    run_name: str
    max_repeats: int
    kwargs: Dict[str, str]
    prepared: Any

    def __init__(
        self,
        wd: Path,
        project_name: str,
        experiment_name: str,
        run_name: str,
        max_repeats: int,
        rest: List[str],
    ):
        if not experiment_name or experiment_name == "":
            raise ValueError("Experiment name cannot be empty")

        if not project_name or project_name == "":
            raise ValueError("Project name cannot be empty")

        if not run_name or run_name == "":
            raise ValueError("Run name cannot be empty")

        if not max_repeats or max_repeats < -1:
            raise ValueError("Max repeats cannot be none or less than -1")

        self.wd = wd
        self.experiment_name = experiment_name
        self.project_name = project_name
        self.run_name = run_name
        self.max_repeats = max_repeats
        self.kwargs = self._parse(rest)
        self._find_route()
        self.eval_params()
        self.apply_train()

    def _parse(self, rest: List[str]):
        kwargs = {}
        for arg in rest:
            if "=" not in arg:
                continue
            key, value = arg.split("=")
            kwargs[key] = value
        return kwargs

    def _find_route(self):
        _source_experiments(self.wd / "src")

        if self.experiment_name not in _experiments:
            raise ValueError(f"Experiment {self.experiment_name} not found")

        experiment = _experiments[self.experiment_name]
        self.experiment = experiment

    def eval_params(self):
        params = parse_kwargs_to_params(self.experiment.params, self.kwargs)
        setattr(params, "experiment_name", self.experiment_name)
        setattr(params, "project_name", self.project_name)
        setattr(params, "run_name", self.run_name)
        setattr(params, "rank", int(os.environ.get("RANK", 0)))
        setattr(params, "world_size", int(os.environ.get("WORLD_SIZE", 1)))
        setattr(params, "local_rank", int(os.environ.get("LOCAL_RANK", 0)))

        self.prepared = params

    def apply_train(self):
        self.experiment.train(self.prepared)



#This code is from openai baseline
#https://github.com/openai/baselines/tree/master/baselines/common/vec_env

import numpy as np
from multiprocessing import Process, Pipe

def worker(remote, parent_remote, env_fn_wrapper):
    parent_remote.close()
    env = env_fn_wrapper.x()
    while True:
        cmd, data = remote.recv()
        if cmd == 'step':
            ob, reward, done, info = env.step(data)
            if done:
                ob = env.reset()
            remote.send((ob, reward, done, info))
        elif cmd == 'reset':
            ob = env.reset()
            remote.send(ob)
        elif cmd == 'reset_task':
            ob = env.reset_task()
            remote.send(ob)
        elif cmd == 'close':
            remote.close()
            break
        elif cmd == 'get_spaces':
            remote.send((env.observation_space, env.action_space))
        else:
            raise NotImplementedError

class VecEnv(object):
    """
    An abstract asynchronous, vectorized environment.
    """
    def __init__(self, num_envs, observation_space, action_space):
        self.num_envs = num_envs
        self.observation_space = observation_space
        self.action_space = action_space

    def reset(self):
        """
        Reset all the environments and return an array of
        observations, or a tuple of observation arrays.
        If step_async is still doing work, that work will
        be cancelled and step_wait() should not be called
        until step_async() is invoked again.
        """
        pass

    def step_async(self, actions):
        """
        Tell all the environments to start taking a step
        with the given actions.
        Call step_wait() to get the results of the step.
        You should not call this if a step_async run is
        already pending.
        """
        pass

    def step_wait(self):
        """
        Wait for the step taken with step_async().
        Returns (obs, rews, dones, infos):
         - obs: an array of observations, or a tuple of
                arrays of observations.
         - rews: an array of rewards
         - dones: an array of "episode done" booleans
         - infos: a sequence of info objects
        """
        pass

    def close(self):
        """
        Clean up the environments' resources.
        """
        pass

    def step(self, actions):
        self.step_async(actions)
        return self.step_wait()

    
class CloudpickleWrapper(object):
    """
    Uses cloudpickle to serialize contents (otherwise multiprocessing tries to use pickle)
    """
    def __init__(self, x):
        self.x = x
    def __getstate__(self):
        import cloudpickle
        return cloudpickle.dumps(self.x)
    def __setstate__(self, ob):
        import pickle
        self.x = pickle.loads(ob)

        
class SubprocVecEnv(VecEnv):
    def __init__(self, env_fns, spaces=None):
        """
        envs: list of gym environments to run in subprocesses
        """
        self.waiting = False
        self.closed = False
        nenvs = len(env_fns)
        self.nenvs = nenvs
        self.remotes, self.work_remotes = zip(*[Pipe() for _ in range(nenvs)])
        self.ps = [Process(target=worker, args=(work_remote, remote, CloudpickleWrapper(env_fn)))
            for (work_remote, remote, env_fn) in zip(self.work_remotes, self.remotes, env_fns)]
        for p in self.ps:
            p.daemon = True # if the main process crashes, we should not cause things to hang
            p.start()
        for remote in self.work_remotes:
            remote.close()

        self.remotes[0].send(('get_spaces', None))
        observation_space, action_space = self.remotes[0].recv()
        VecEnv.__init__(self, len(env_fns), observation_space, action_space)

    def step_async(self, actions):
        for remote, action in zip(self.remotes, actions):
            remote.send(('step', action))
        self.waiting = True

    def step_wait(self):
        results = [remote.recv() for remote in self.remotes]
        self.waiting = False
        obs, rews, dones, infos = zip(*results)
        return np.stack(obs), np.stack(rews), np.stack(dones), infos

    def reset(self):
        for remote in self.remotes:
            remote.send(('reset', None))
        return np.stack([remote.recv() for remote in self.remotes])

    def reset_task(self):
        for remote in self.remotes:
            remote.send(('reset_task', None))
        return np.stack([remote.recv() for remote in self.remotes])

    def close(self):
        if self.closed:
            return
        if self.waiting:
            for remote in self.remotes:            
                remote.recv()
        for remote in self.remotes:
            remote.send(('close', None))
        for p in self.ps:
            p.join()
            self.closed = True
            
    def __len__(self):
        return self.nenvs


import torch.optim as optim
from torch.optim.lr_scheduler import StepLR

from higgsfield.llama import Llama
from higgsfield.loaders import LlamaLoader 
from higgsfield.checkpoint import Checkpoint
from higgsfield.training import  clip_grad_norm
from higgsfield.experiment import experiment, param

from src.dataset import AlpacaDataset

@experiment("alpaca_bf16")
@param("size", options=["7b", "13b", "70b"])
@param("num_epochs", default=1, description="Number of epochs")
def train(params):
    
    if params.size == "7b":
        model_name = "meta-llama/Llama-2-7b-hf"
    elif params.size == "13b":
        model_name = "meta-llama/Llama-2-13b-hf"
    elif params.size == "70b":
        model_name = "meta-llama/Llama-2-70b-hf"
    
    model = Llama(
        model_name=model_name,
        zero_stage=3,
        cpu_init_rank0=True,
        fast_attn=False,
        precision="bf16",
        cpu_offload=False,
    )
    
    optimizer = optim.AdamW(
        model.parameters(),
        lr=1e-5,
        weight_decay=0.0,
    )
    
    lr_scheduler = StepLR(
        optimizer,
        step_size=1,
        gamma=0.85,
    )
    
    # ~/.cache/{project-name}/experiments/{experiment_name}/{run_name}/
    checkpoint = Checkpoint(
        model,
        optimizer,
        lr_scheduler,
    )
    
    dataset_name = "tatsu-lab/alpaca"
    dataset = AlpacaDataset(dataset_name, split="train")
    
    train_loader = LlamaLoader(
        dataset,
        max_sequence_length=2048,
        batch_size_per_gpu=1,
    )
    
    for epoch in range(params.num_epochs):
        for i, batch in enumerate(train_loader):
            
            optimizer.zero_grad()
            loss = model(batch)
            
            if params.rank == 0:
                print("Loss: ", loss)
                
            loss.backward()
            
            clip_grad_norm(1.0, model, optimizer)
            optimizer.step()

            if i % 30 == 0 or i == len(train_loader) - 1:
                checkpoint.save(epoch, i)
            
        lr_scheduler.step()
        
    model.save_huggingface_model("my-alpaca")
    
    



import pathlib


class _ProjectCachePath:
    project_name: str
    path: pathlib.Path
    metadata_file: str
    _init_path: pathlib.Path

    def _init(
        self,
        project_name: str,
        metadata_file: str,
        init_path: pathlib.Path,
        verbose: bool = True,
    ):
        self.project_name = project_name
        self.metadata_file = metadata_file
        self._init_path = init_path
        self.verbose = verbose
        self.path = self._mkdir()

    def experiment_path(self, experiment_name: str) -> "ExperimentPath":
        return ExperimentPath(self, experiment_name)

    def _mkdir(self) -> pathlib.Path:
        home = self._init_path

        path = home / f".cache/{self.project_name}"

        try:
            (path / "experiments").mkdir(exist_ok=True, parents=True)
        except Exception as e:
            if self.verbose:
                print("this error shouldn't have been thrown")
                print(f"error creating path {path}")
                print(e)

        return path

    def metadata_path(self) -> pathlib.Path:
        return (self.path / "experiments") / self.metadata_file


class ProjectCachePath(_ProjectCachePath):
    def __init__(
        self,
        project_name: str,
        metadata_file: str = "metadata.json",
    ):
        home = pathlib.Path.home()
        self._init(project_name, metadata_file, home)


class ExperimentPath:
    project_path: _ProjectCachePath
    experiment_name: str
    path: pathlib.Path
    metadata_file: str

    def __init__(self, project_path: _ProjectCachePath, experiment_name: str):
        self.project_path = project_path
        self.experiment_name = experiment_name

        self.path = self._mkdir()

    def run_path(self, run_name: str) -> "RunPath":
        return RunPath(self, run_name)

    def _mkdir(self) -> pathlib.Path:
        path = self.project_path.path / f"experiments/{self.experiment_name}"

        try:
            path.mkdir(exist_ok=True, parents=True)
        except Exception as e:
            if self.project_path.verbose:
                print("this error shouldn't have been thrown")
                print(f"error creating path {path}")
                print(e)
        return path


class RunPath:
    experiment_path: ExperimentPath
    run_name: str

    def __init__(self, experiment_path: ExperimentPath, run_name: str):
        self.experiment_path = experiment_path
        self.run_name = run_name

    def checkpoint_path(self) -> pathlib.Path:
        path = self.experiment_path.path / f"checkpoints/{self.run_name}"

        try:
            path.mkdir(exist_ok=True, parents=True)
        except Exception as e:
            if self.experiment_path.project_path.verbose:
                print("this error shouldn't have been thrown")
                print(f"error creating path {path}")
                print(e)
        return path

    def sharded_checkpoint_path(self) -> pathlib.Path:
        path = self.experiment_path.path / f"sharded-checkpoints/{self.run_name}"

        try:
            path.mkdir(exist_ok=True, parents=True)
        except Exception as e:
            if self.experiment_path.project_path.verbose:
                print("this error shouldn't have been thrown")
                print(f"error creating path {path}")
                print(e)
        return path

    def lr_scheduler_path(self) -> pathlib.Path:
        path = self.experiment_path.path / f"lr-schedulers/{self.run_name}"

        try:
            path.mkdir(exist_ok=True, parents=True)
        except Exception as e:
            if self.experiment_path.project_path.verbose:
                print("this error shouldn't have been thrown")
                print(f"error creating path {path}")
                print(e)

        return path


def working_directory() -> pathlib.Path:
    return pathlib.Path.cwd()



from pathlib import Path
from typing import List

import click

from .util import wd_path, check_name, parse_origin_link_or_else
from .launch import Launch

from .cfg import AppConfig
from .experiment.builder import DeployBuilder, KillBuilder, build_all_experiment_actions

from .ci import cli as ci_cli
from higgsfield.internal.init import init

import os


def setup_environ_flags(rank):
    """Environment flags for debugging purposes"""

    os.environ["TORCH_SHOW_CPP_STACKTRACES"] = str(1)
    os.environ["NCCL_ASYNC_ERROR_HANDLING"] = str(1)


def setup(seed):
    import torch
    import torch.distributed as dist

    torch.cuda.manual_seed(seed)
    torch.manual_seed(seed)

    dist.init_process_group(backend="nccl")

    local_rank = int(os.environ["LOCAL_RANK"])
    rank = int(os.environ["RANK"])
    world_size = int(os.environ["WORLD_SIZE"])

    if dist.is_initialized():
        torch.cuda.set_device(local_rank)
        torch.cuda.empty_cache()
        setup_environ_flags(rank)


@click.command("run")
@click.option("--experiment_name", type=str, help="experiment name")
@click.option("--run_name", type=str, help="run name")
@click.option("--max_repeats", type=int, help="max repeats")
@click.argument("extra_args", nargs=-1)
def run_experiment(
    experiment_name: str,
    run_name: str,
    max_repeats: int,
    extra_args: List[str],
):  
    wd = wd_path()
    app_config = AppConfig.from_path(wd)

    os.environ["PROJECT_NAME"] = app_config.name
    os.environ["EXPERIMENT_NAME"] = experiment_name
    os.environ["RUN_NAME"] = run_name
    setup(42)
    Launch(wd, app_config.name, experiment_name, run_name, max_repeats, extra_args)


@click.command("init")
@click.argument("project_name", type=str, required=True)
def init_cmd(project_name: str):
    print(
        "Initializing {} project at:\n{}/{}".format(
            project_name, wd_path(), project_name
        )
    )
    init(wd_path(), check_name(project_name))


@click.command("build-experiments")
def build_experiments():
    wd = wd_path()
    
    app_config = AppConfig.from_path(wd)
    
    origin_url = app_config.get_git_origin_url(wd)
    
    if origin_url is None:
        raise ValueError("Have you pushed your project to github?")

    origin_url = parse_origin_link_or_else(origin_url)

    if origin_url is None:
        raise ValueError("Please use ssh or https url for github repo.")

    app_config.github_repo_url = origin_url

    app_config.set_git_origin_url(wd)
    
    DeployBuilder(app_config, wd).generate()
    KillBuilder(app_config, wd).generate()
    build_all_experiment_actions(wd, app_config)


@click.command("show-deploy-key")
def show_deploy_key():
    wd = wd_path()
    app_config = AppConfig.from_path(wd)
    deploy_key_path = (
        Path.home() / ".ssh" / "higgsfield" / f"{app_config.name}-github-deploy.key.pub"
    )
    if not deploy_key_path.exists():
        raise ValueError("No deploy key found, file an issue on github")

    with deploy_key_path as f:
        click.echo(f.read_text() + "\n")


@click.group("ci")
def ci():
    pass


ci.add_command(ci_cli.proc_per_node)
ci.add_command(ci_cli.ssh_details)
ci.add_command(ci_cli.decode_secrets)



from .dataset import CompletionDataset

def chat_to_prompt(chat):
    joined = []
    for message in chat:
        joined.append(f"###{message['role'].upper()}: {message['content']}")
    
    prompt = "\n".join(joined)
    prompt += "\n###ASSISTANT: "
    
    return prompt

class ChatCompletionDataset(CompletionDataset):
    ''' OpenAI's api format:
    chats = [
        [
          {"role": "system", "content": "You are a helpful assistant."},
          {"role": "user", "content": "Who won the world series in 2020?"},
          {"role": "assistant", "content": "The Los Angeles Dodgers won the World Series in 2020."},
          {"role": "user", "content": "Where was it played?"}
        ],
    ]
    '''
    def __init__(self, chats, chat_to_prompt=chat_to_prompt):
        self.chat_to_prompt = chat_to_prompt
        
        self.chats = chats
        
        items = []
        for chat in self.chats:
            current_chat = []
            last_user = False
            
            for message in chat:
                if message["role"] == "system":
                    current_chat.append(message)
                    
                elif message["role"] == "user":
                    last_user = True
                    current_chat.append(message)
                    
                elif message["role"] == "assistant":
                    if last_user:
                        items.append([
                            [c for c in current_chat], message["content"]
                        ])
                    current_chat.append(message)
                    
        self.items = items
        
    def __len__(self):
        return len(self.items)
        
    def __getitem__(self, idx):
        chat, completion = self.items[idx]
        prompt = self.chat_to_prompt(chat)
        
        return {
            "prompt": prompt,
            "completion": completion
        }


from typing import Any, List, Tuple, Dict, Optional, Type
from yaml import safe_dump
from higgsfield.internal.util import check_name


def build_run_params(params: List["Param"]) -> str:
    return " ".join(param.as_run_param() for param in params)


def build_gh_action_inputs(params: List["Param"]) -> List[str]:
    return [param.as_github_action() for param in params]


_arg_type_set = {
    int,
    float,
    str,
    bool,
}


def reset_values(values: Dict[str, Any]):
    if "name" not in values:
        raise ValueError("Field name is required in Param")
    if "required" not in values:
        values["required"] = False

    return values


class Param:
    name: str
    default: Optional[Any] = None
    description: Optional[str] = None
    required: bool = False
    type: Optional[Type] = None
    options: Optional[Tuple[Any, ...]] = None
    
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

    @classmethod
    def from_values(cls, **values) -> "Param":
        values = reset_values(values)
        values["name"] = check_name(values.get("name") or "")
        arg_type = values.get("type")
        default = values.get("default")

        if arg_type is None:
            if default is None:
                arg_type = str
                values["type"] = arg_type
            else:
                arg_type = type(default)
                values["type"] = arg_type

        if arg_type not in _arg_type_set:
            raise ValueError(
                f"Param type {arg_type} not supported."
                + "Only primitive {_arg_type_map.keys()} are supported"
            )

        try:
            if default is not None:
                values["default"] = arg_type(default)
        except Exception as e:
            raise ValueError(
                f"Param default {default} cannot be converted to type {arg_type}"
            ) from e

        if options := values.get("options"):
            values["options"] = [arg_type(opt) for opt in options]
            if default is None:
                values["default"] = values["options"][0]
                default = values["default"]

            if default is not None and default not in values["options"]:
                raise ValueError(f"Param default {default} not in options {options}")

        return Param(**values)

    def check(self, value: str):
        try:
            if self.type is None:
                self.type = str

            value = self.type(value)
        except Exception as e:
            raise ValueError(
                f"Param value {value} cannot be converted to type {self.type}"
            ) from e

        if options := self.options:
            if value not in options:
                raise ValueError(f"Param value {value} not in options {options}")

    def as_github_action(self) -> str:
        indent = "        "
        to_join = [f"{self.name}:"]
        if self.description:
            # TODO: fix that yaml.safe_dump with some proper encoder, string esc etc.
            to_join.append(
                f"{indent}description: {remove_trailing_yaml(safe_dump(self.description))}"
            )
        to_join.append(f"{indent}required: {self.required}")
        if self.default:
            d = self.default
            if type(self.default) == str:
                d = remove_trailing_yaml(safe_dump(self.default))
            to_join.append(f"{indent}default: {d}")
        if self.options:
            to_join.append(f"{indent}options: {list(self.options)}")
            to_join.append(f"{indent}type: choice")
        if self.type == bool:
            to_join.append(f"{indent}type: boolean")
        return "\n".join(to_join)

    def as_run_param(self) -> str:
        # field_name="value"
        return f'{pfx}{self.name}="{wrap_brackets("github.event.inputs." + self.name)}"'

    class Config:
        frozen = True


def wrap_brackets(s: str) -> str:
    left = "${{ "
    right = " }}"
    return left + s + right


def remove_trailing_yaml(s: str) -> str:
    trailing_yaml = "\n...\n"
    if s.endswith(trailing_yaml):
        return s[: -len(trailing_yaml)]
    return s


pfx = "hf_action_"


class _ToSet:
    param: Param
    value: Any

    def __init__(self, param: Param, value: Optional[Any] = None):
        self.param = param
        self.value = value


class ArgParams:
    pass


def parse_kwargs_to_params(
    params: List[Param],
    kwargs: Dict[str, str],
):
    keys = (key[len(pfx) :] for key in kwargs if key.startswith(pfx))

    fields = {param.name: param for param in params}

    # remove keys that are not in subclass_fields
    keys = [key for key in keys if key in fields]

    # get values from kwargs
    values = [kwargs[pfx + key] for key in keys]

    k_v = dict(zip(keys, values))

    params_to_set = {key: _ToSet(param=param) for key, param in fields.items()}

    for key, to_set in params_to_set.items():
        if key in k_v:
            v = k_v[key]
            to_set.param.check(v)
            to_set.value = v
        elif to_set.param.required and to_set.param.default is None:
            raise ValueError(f"Required argument {key} not provided")
        else:
            to_set.value = to_set.param.default

        params_to_set[key] = to_set

    prepare = ArgParams()
    for key, to_set in params_to_set.items():
        setattr(prepare, key, to_set.param.type(to_set.value))

    return prepare



import torch
from transformers import (
    LlamaConfig,
    LlamaForCausalLM,
)
from higgsfield.checkpoint import fsdp_model_state_dict_rank0

def load_llama_from_config(model_name):
    config = LlamaConfig.from_pretrained(model_name)
    model = LlamaForCausalLM(config)
    return model

def load_llama_from_checkpoint(model_name, checkpoint_path):
    model = load_llama_from_config(model_name)
    state_dict = torch.load(checkpoint_path)
    model.load_state_dict(state_dict)
    return model


from typing import List, Optional

import os
import dotenv

from higgsfield.internal.cfg import AppConfig

from .params import Param, build_gh_action_inputs, build_run_params
from jinja2 import Environment, FileSystemLoader, Template
from higgsfield.internal.util import templates_path
from pathlib import Path
from importlib.machinery import SourceFileLoader
from .decorator import ExperimentDecorator
from higgsfield.internal.experiment.ast_parser import parse_experiments

header = """# THIS FILE WAS GENERATED BY HIGGSFIELD. 
# DO NOT EDIT. 
# IF YOUR WORKFLOW DOESN'T WORK, CREATE AN ISSUE.
"""


class ActionBuilder:
    app_config: AppConfig
    wf_dir: Path
    template: Template
    template_name: str

    def __init__(self, app_config: AppConfig, project_path: Path):
        environment = Environment(loader=FileSystemLoader(templates_path()))
        self.template = environment.get_template(self.template_name)

        self.app_config = app_config

        wf_dir = project_path / ".github" / "workflows"
        if not wf_dir.exists():
            wf_dir.mkdir(parents=True, exist_ok=True)
        elif not wf_dir.is_dir():
            raise ValueError(f"{wf_dir} is not a directory, delete it and try again")

        self.wf_dir = wf_dir


class KillBuilder(ActionBuilder):
    template_name: str = "kill_action.j2"

    def generate(self):
        (self.wf_dir / "kill.yml").write_text(
            self.template.render(
                header=header,
                project_name=self.app_config.name,
            )
        )

        print("Updated kill action")


def as_keyed_repo_url(repo_url: Optional[str], project_name: str) -> str:
    if repo_url is None:
        raise ValueError("Did you add git remote origin? / push to github?")

    # get the index of "/" after "github.com"
    # and replace it with "-project_name:"
    return repo_url.replace("github.com:", f"github.com-{project_name}:")


def insert_env_line(keys: List[str], indent: str) -> str:
    lines = []
    for key in keys:
        line = indent + "echo " + key + '="${{ secrets.' + key + ' }}" >> env'
        lines.append(line)

    return "\n".join(lines)


def env_keys_as_action(path: Path, indent: str) -> str:
    keys = list(dotenv.dotenv_values(path).keys())
    keys.pop(keys.index("SSH_KEY"))

    return insert_env_line(keys, indent)


echo_indent = "          "


class DeployBuilder(ActionBuilder):
    template_name = "deploy_action.j2"

    def generate(self):
        (self.wf_dir / "deploy.yml").write_text(
            self.template.render(
                header=header,
                project_name=self.app_config.name,
                keyed_repo_url=as_keyed_repo_url(
                    self.app_config.github_repo_url, self.app_config.name
                ),
                env_gen=env_keys_as_action(
                    self.wf_dir.parent.parent / "env", echo_indent
                ),
            )
        )

        print("Updated deploy action")


class ExperimentBuilder(ActionBuilder):
    template_name = "experiment_action.j2"

    def generate(self, experiment_name: str, params: List[Param]):
        (self.wf_dir / f"run_{experiment_name}.yml").write_text(
            self.template.render(
                header=header,
                experiment_name=experiment_name,
                project_name=self.app_config.name,
                params=build_gh_action_inputs(params),
                rest=build_run_params(params),
                env_gen=env_keys_as_action(
                    self.wf_dir.parent.parent / "env", echo_indent
                ),
            )
        )
        

        print("Updated experiment action", experiment_name)


def _source_experiments(base_path: Path):
    """
    Only used inside docker to inject experiments into the module.
    Do not use outside of docker. But if you do, you will have to have
    the same dependencies (aka environment) as the docker container.
    """
    for file in base_path.glob("**/*.py"):
        module_name = os.path.basename(file).split(".py")[0].split(".py")[0].replace(" ", "_").replace("-", "_")
        SourceFileLoader(module_name, str(file)).load_module()


def build_all_experiment_actions(wd_path: Path, app_config: AppConfig):
    """
    Builds all experiment actions

    Root path should be the root of the project and must contain the src folder under which everything is defined.
    Project name should be the name of the project.

    It will parse ast of all files under src folder and search for experiments.
    Then it will build actions for each experiment and save them under .github/workflow folder.
    Deletes all actions that have name prefix run_experiment_ and header inside if not needed.
    """
    exp_params_pairs = []
    for file in (wd_path / "src").glob("**/*.py"):
        exp_params_pairs.extend(parse_experiments(str(file.resolve())))

    if len(exp_params_pairs) == 0:
        print("No experiments found")
        return

    experiments = ExperimentDecorator.from_ast(exp_params_pairs)
    actions_folder = wd_path / ".github" / "workflows"
    actions_folder.mkdir(parents=True, exist_ok=True)

    # list all files that have name prefix run_experiment_ and header inside
    for i in actions_folder.glob("run_*.yml"):
        if i.read_text().startswith(header):
            i.unlink()

    for experiment_name, experiment in experiments.items():
        ExperimentBuilder(app_config, wd_path).generate(
            experiment_name, experiment.params
        )



from pathlib import Path
import os
import io

from cryptography.hazmat.primitives import serialization, asymmetric
import paramiko
from jinja2 import Environment, FileSystemLoader
from .util import templates_path, ROOT_DIR

from typing import Tuple

templates = Environment(loader=FileSystemLoader(templates_path()))


def init(wd: Path, project_name: str):
    if project_name is None or project_name == "":
        raise ValueError("Project name cannot be empty")
    project_path = wd / project_name
    if project_path.exists():
        raise ValueError(f"Project {project_name} already exists")

    project_path.mkdir(parents=True, exist_ok=True)
    (project_path / "src").mkdir(parents=True, exist_ok=True)

    config_path = project_path / "src" / "config.py"
    if config_path.exists():
        raise ValueError(f"Config file {config_path} already exists")

    readme_path = project_path / "README.md"
    if readme_path.exists():
        raise ValueError(f"README.md file {readme_path} already exists")

    source_path = Path(ROOT_DIR) / "static" / "project"
    fileset = [
        ".gitignore",
        "env",
        "Dockerfile",
        "src/alpaca_bf16.py",
        "src/alpaca_fp16.py",
        "src/dataset.py",
        "requirements.txt",
    ]

    for file in fileset:
        (project_path / file).write_bytes((source_path / file).read_bytes())

    config_path.write_text(
        templates.get_template("config_py.j2").render(project_name=project_name)
    )

    readme_path.write_text(
        templates.get_template("README_md.j2").render(project_name=project_name)
    )

    hf_deploy_ssh_folder = Path.home() / ".ssh/higgsfield/"
    hf_deploy_ssh_folder.mkdir(parents=True, exist_ok=True)
    priv, pub, _ = generate_deploy_keys()
    (hf_deploy_ssh_folder / f"{project_name}-github-deploy.key").write_bytes(priv)

    # set permissions to 400
    os.system(f"chmod 600 {hf_deploy_ssh_folder / f'{project_name}-github-deploy.key'}")
    public_key_path = hf_deploy_ssh_folder / f"{project_name}-github-deploy.key.pub"
    public_key_path.write_text(pub.decode() + "\n")

    # set permissions to 400
    os.system(
        f"chmod 600 {hf_deploy_ssh_folder / f'{project_name}-github-deploy.key.pub'}"
    )


def generate_deploy_keys() -> Tuple[bytes, bytes, paramiko.Ed25519Key]:
    c_ed25519key = asymmetric.ed25519.Ed25519PrivateKey.generate()  # type: ignore
    privpem = c_ed25519key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.OpenSSH,
        encryption_algorithm=serialization.NoEncryption(),
    )
    priv_obj = io.StringIO(privpem.decode())
    p_ed25519key = paramiko.Ed25519Key.from_private_key(priv_obj)

    pub = c_ed25519key.public_key()
    openssh_pub = pub.public_bytes(
        encoding=serialization.Encoding.OpenSSH,
        format=serialization.PublicFormat.OpenSSH,
    )

    return privpem, openssh_pub, p_ed25519key



import copy
import torch
from torch.utils.data import Dataset

class CompletionDataset(Dataset):
    '''
    def __getitem__(self, idx):
        ...
        return {
            "prompt": prompt,
            "completion": completion,
        }
    '''
    pass

class LMDataset(Dataset):
    '''
    def __getitem__(self, idx):
        ...
        return "whatever sequence you want to return as a string"
    '''
    pass

class TorchMultiTurnDataset(Dataset):
    def __init__(self, dataset, tokenizer, max_sequence_length):
        self.dataset             = dataset
        self.tokenizer           = tokenizer
        self.max_sequence_length = max_sequence_length
    
    def __len__(self):
        return len(self.dataset)
    
    def __getitem__(self, idx):
        items = self.dataset[idx]
        
        IGNORE_INDEX = -100
        
        multi_labels  = []
        multi_example = []
        for item in items:
            prompt     = item["prompt"]
            completion = item["completion"]
            
            example = prompt + completion
            prompt = torch.tensor(
                self.tokenizer.encode(prompt), dtype=torch.int64
            )
            example = self.tokenizer.encode(example)
            example.append(self.tokenizer.eos_token_id)
            example = torch.tensor(
                example, dtype=torch.int64
            )
                
            labels = copy.deepcopy(example)
            labels[: len(prompt)] = -1
            
            multi_example.append(example)
            multi_labels.append(labels)
            
        example = torch.cat(multi_example)
        labels  = torch.cat(multi_labels)
        
        padding = self.max_sequence_length - example.shape[0]
        if padding > 0:
            example = torch.cat((example, torch.zeros(padding, dtype=torch.int64) - 1))
            labels  = torch.cat((labels,  torch.zeros(padding, dtype=torch.int64) - 1))
        elif padding < 0:
            example = example[: self.max_sequence_length]
            labels  = labels[:  self.max_sequence_length]
        
        label_mask = labels.ge(0)
        labels[~label_mask] = IGNORE_INDEX
        
        example_mask = example.ge(0)
        example[~example_mask] = 0
        example_mask = example_mask.float()
        
        return {
            "input_ids": example,
            "labels": labels,
            "attention_mask": example_mask,
        }

class TorchCompletionDataset(Dataset):
    def __init__(self, dataset, tokenizer, max_sequence_length):
        self.dataset             = dataset
        self.tokenizer           = tokenizer
        self.max_sequence_length = max_sequence_length
        
    def __len__(self):
        return len(self.dataset)
        
    def __getitem__(self, idx):
        item = self.dataset[idx]
        
        prompt = item["prompt"]
        completion = item["completion"]
        
        IGNORE_INDEX = -100  # The default setting in CrossEntropyLoss
        
        example = prompt + completion
        prompt = torch.tensor(
            self.tokenizer.encode(prompt), dtype=torch.int64
        )
        example = self.tokenizer.encode(example)
        example.append(self.tokenizer.eos_token_id)
        example = torch.tensor(
            example, dtype=torch.int64
        )
        padding = self.max_sequence_length - example.shape[0]
        if padding > 0:
            example = torch.cat((example, torch.zeros(padding, dtype=torch.int64) - 1))
        elif padding < 0:
            example = example[: self.max_sequence_length]
            
        labels = copy.deepcopy(example)
        labels[: len(prompt)] = -1
        label_mask = labels.ge(0)
        labels[~label_mask] = IGNORE_INDEX
        
        example_mask = example.ge(0)
        example[~example_mask] = 0
        example_mask = example_mask.float()
        
        return {
            "input_ids": example,
            "labels": labels,
            "attention_mask":example_mask,
        }

class TorchLMDataset(Dataset):
    def __init__(self, dataset, tokenizer, max_sequence_length):
        self.dataset   = dataset
        self.tokenizer = tokenizer
        self.max_sequence_length = max_sequence_length
       
    def __len__(self):
        return len(self.dataset)
    
    def __getitem__(self, idx):
        x = self.dataset[idx]
        
        IGNORE_INDEX = -100
        
        example = self.tokenizer.encode(x, add_special_tokens=False)
        example = torch.tensor(
            example, dtype=torch.int64
        )
        
        padding = self.max_sequence_length - example.shape[0]
        if padding > 0:
            example = torch.cat((example, torch.zeros(padding, dtype=torch.int64) - 1))
        elif padding < 0:
            example = example[: self.max_sequence_length]
       
        labels = copy.deepcopy(example)
        example_mask = example.ge(0)
        label_mask = labels.ge(0)
        example[~example_mask] = 0
        labels[~label_mask] = IGNORE_INDEX
        example_mask = example_mask.float()
        label_mask = label_mask.float()       
        
        return {
            "input_ids": example,
            "labels": labels,
            "attention_mask":example_mask,
        }
        
        


import os
import functools
from pathlib import Path

import torch
import torch.distributed as dist

from torch.distributed.fsdp.fully_sharded_data_parallel import (
    FullyShardedDataParallel as FSDP,
    CPUOffload,
)

from torch.distributed.fsdp import (
    MixedPrecision,
    ShardingStrategy,
)
from torch.distributed.fsdp.wrap import (
    transformer_auto_wrap_policy,
)
from torch.distributed.algorithms._checkpoint.checkpoint_wrapper import (
    checkpoint_wrapper,
    CheckpointImpl,
    apply_activation_checkpointing,
)

from transformers import (
    LlamaForCausalLM,
    LlamaTokenizer,
    LlamaConfig,
    default_data_collator,
)
from transformers.models.llama.modeling_llama import LlamaDecoderLayer
from optimum.bettertransformer import BetterTransformer

from higgsfield.checkpoint.fsdp_checkpoint import (
    save_distributed_model_rank0,
    fsdp_model_state_dict_rank0,
)

from .llama_utils import (
    load_llama_from_checkpoint,
    load_llama_from_config,
)

class Llama(FSDP):
    def __init__(
        self,
        model_name,
        checkpoint_path=None,
        zero_stage=3,
        fast_attn=False,
        precision="bf16",
        cpu_init_rank0=False,
        cpu_offload=False,
    ):
        
        rank = dist.get_rank()
        
        if not checkpoint_path:
            if cpu_init_rank0: 
                if rank == 0:
                    model = LlamaForCausalLM.from_pretrained(model_name)
                else:
                    llama_config = LlamaConfig.from_pretrained(model_name)
                        
                    with torch.device('meta'):
                        model = LlamaForCausalLM(llama_config)
            else:
                model = LlamaForCausalLM.from_pretrained(model_name)
        else:
            if not cpu_init_rank0:
                print("Ignoring cpu_init_rank0=False while loading model from checkpoint path")
                cpu_init_rank0 = True
            
            if rank == 0:
                model = load_llama_from_checkpoint(model_name, checkpoint_path)
                print("LOADED FROM CHECKPOINT")
                
            else:
                llama_config = LlamaConfig.from_pretrained(model_name)
                
                with torch.device('meta'):
                    model = LlamaForCausalLM(llama_config)
            
        if fast_attn:
            #raise NotImplementedError("Fast attention is not supported yet")
            model = BetterTransformer.transform(model)
        
        fpSixteen = MixedPrecision(
            param_dtype=torch.float16,
            reduce_dtype=torch.float16,
            buffer_dtype=torch.float16,
        )

        bfSixteen_mixed = MixedPrecision(
            param_dtype=torch.float32,
            reduce_dtype=torch.bfloat16,
            buffer_dtype=torch.bfloat16,
        )
        
        pure_bf16 = False
        if precision == "fp16":
            mixed_precision_policy = fpSixteen
            
        elif precision == "bf16":
            mixed_precision_policy = None
            pure_bf16 = True
            
        elif precision == "bf16_mixed":
            mixed_precision_policy = bfSixteen_mixed

        else:
            mixed_precision_policy = None
            
        if pure_bf16:
            model.to(torch.bfloat16) 

        wrapping_policy = functools.partial(
            transformer_auto_wrap_policy,
            transformer_layer_cls={
                LlamaDecoderLayer,
            }
        )
        
        if zero_stage == 0:
            sharding_strategy = ShardingStrategy.NO_SHARD
        
        elif zero_stage == 1:
            raise NotImplementedError("stage 1 is not supported. Only 0 2 3")
            
        elif zero_stage == 2:
            sharding_strategy = ShardingStrategy.SHARD_GRAD_OP
            
        elif zero_stage == 3:
            sharding_strategy = ShardingStrategy.FULL_SHARD
        else:
            raise NotImplementedError("stage can be only 0 2 3")

        if cpu_init_rank0 and rank != 0:
            param_init_fn = lambda module: module.to_empty(
                device=torch.device('cuda'),
                recurse=False,
            )
        else:
            param_init_fn = None
            
        if cpu_offload:
            cpu_offload = CPUOffload(offload_params=True)
        else:
            cpu_offload = None

        super().__init__(
            model,
            auto_wrap_policy=wrapping_policy,
            cpu_offload=cpu_offload,
            mixed_precision=mixed_precision_policy,
            sharding_strategy=sharding_strategy,
            device_id=torch.cuda.current_device(),
            limit_all_gathers=True,
            sync_module_states=cpu_init_rank0,
            param_init_fn=param_init_fn,
        )
        
        non_reentrant_wrapper = functools.partial(
            checkpoint_wrapper,
            checkpoint_impl=CheckpointImpl.NO_REENTRANT,
        )

        check_fn = lambda submodule: isinstance(submodule, LlamaDecoderLayer)

        apply_activation_checkpointing(
            self,
            checkpoint_wrapper_fn=non_reentrant_wrapper,
            check_fn=check_fn,
        )
        
        fsdp = True
        self.precision = precision
        self.fsdp = fsdp
        self.model_name = model_name
    
    def __call__(self, batch):
        local_rank = int(os.environ["LOCAL_RANK"])
        
        for key in batch.keys():
            batch[key] = batch[key].to(local_rank)
            
        if self.precision == "fp16":
            with torch.cuda.amp.autocast(): 
                loss = super().__call__(**batch).loss
        else:
            loss = super().__call__(**batch).loss
            
        return loss
    
    def save_model(self, save_path):
        '''
            Save model's weight to master node 
                ~/.cache/higgsfield/{save_path}
        '''
        if "/" == save_path[0]:
            save_path = save_path[1:]
            
        head, tail = os.path.split(save_path)
        
        path = Path.home() / ".cache/higgsfield" / head
        path.mkdir(exist_ok=True, parents=True)
    
        save_distributed_model_rank0(path / tail, self)
        
    def save_huggingface_model(self, save_path):
        '''
            Save model's weight in huggingface format to master node 
                ~/.cache/higgsfield/{save_path}
        '''
        if "/" == save_path[0]:
            save_path = save_path[1:]
            
        head, tail = os.path.split(save_path)
        
        path = Path.home() / ".cache/higgsfield" / head
        path.mkdir(exist_ok=True, parents=True)
        cpu_state = fsdp_model_state_dict_rank0(self)
        
        if dist.get_rank() == 0:
           model = load_llama_from_config(self.model_name)
           model.load_state_dict(cpu_state)
           model.save_pretrained(path / tail)
        
    def push_to_hub(self, repo_id):
        cpu_state = fsdp_model_state_dict_rank0(self)
        
        if dist.get_rank() == 0:
           model = load_llama_from_config(self.model_name)
           model.load_state_dict(cpu_state) 
           model.push_to_hub(repo_id)
        
        
    
class Llama7b(Llama):
    def __init__(
        self,
        checkpoint_path=None,
        zero_stage=3,
        fast_attn=False,
        precision="bf16",
        cpu_init_rank0=False,
        cpu_offload=False,
    ):
        model_name = "meta-llama/Llama-2-7b-hf"
        super(Llama7b, self).__init__(
            model_name,
            checkpoint_path,
            zero_stage,
            fast_attn,
            precision,
            cpu_init_rank0,
            cpu_offload,
        )
       
class Llama13b(Llama):
    def __init__(
        self,
        checkpoint_path=None,
        zero_stage=3,
        fast_attn=False,
        precision="bf16",
        cpu_init_rank0=False,
        cpu_offload=False,
    ):
        model_name = "meta-llama/Llama-2-13b-hf"
        super(Llama13b, self).__init__(
            model_name,
            checkpoint_path,
            zero_stage,
            fast_attn,
            precision,
            cpu_init_rank0,
            cpu_offload,
        )
        
class Llama70b(Llama):
    def __init__(
        self,
        checkpoint_path=None,
        zero_stage=3,
        fast_attn=False,
        precision="bf16",
        cpu_init_rank0=False,
        cpu_offload=False,
    ):
        model_name = "meta-llama/Llama-2-70b-hf"
        super(Llama70b, self).__init__(
            model_name,
            checkpoint_path,
            zero_stage,
            fast_attn,
            precision,
            cpu_init_rank0,
            cpu_offload,
        )


import re
from pathlib import Path

import os

from typing import Optional

regex = re.compile("^[a-zA-Z_][a-zA-Z0-9_]*$")


def check_name(name: str):
    if len(name) < 1 or len(name) > 20:
        raise ValueError("Name must be between 1 and 20 characters long")

    if not regex.match(name):
        raise ValueError("Name must match regex ^[a-zA-Z_][a-zA-Z0-9_]*$")

    return name


def wd_path() -> Path:
    return Path.cwd()


ROOT_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

def templates_path() -> Path:
    return Path(ROOT_DIR) / "static" / "templates"

https_repo_url_pattern = re.compile(
    r"^https\:\/\/github\.com\/[a-zA-Z0-9\-\_]+\/[a-zA-Z0-9\-\_]+\.git$"
)


def match_https_link(link: str) -> bool:
    return https_repo_url_pattern.match(link) is not None


def convert_https_to_ssh(link: str) -> str:
    gh, user, repo = link[8:-4].split("/")
    return f"git@{gh}:{user}/{repo}.git"


def parse_origin_link_or_else(link: str) -> Optional[str]:
    if match_https_link(link):
        return convert_https_to_ssh(link)
    if link.startswith("git@github.com:"):
        return link

    return None



def clip_grad_norm(max_grad_norm, model, optimizer, scaler=None):
    model = model
    
    if scaler:
        scaler.unscale_(optimizer)
        
    if hasattr(optimizer, 'clip_grad_norm'):
        optimizer.clip_grad_norm(max_grad_norm)
        
    elif hasattr(model.model, 'clip_grad_norm_'):
        model.clip_grad_norm_(max_grad_norm)      


from .params import Param
from typing import Callable, Any, List, Tuple, Dict, Optional, Union, Set, Type
from higgsfield.internal.util import check_name
from .ast_parser import Expdec, Paramdec


class InnerWrap:
    param_set: Set[Param]

    def __init__(self, func: Callable[..., None], param_set: Set[Param]):
        self.func = func
        self.param_set = set(param_set)

    def add_param(self, param: Param):
        self.param_set.add(param)


class ExperimentDecorator:
    name: str
    params: List[Param]
    train: Callable[..., None]

    def __init__(self, name: str, *, seed: Optional[int] = None):
        """
        Composable experiment decorator.
        >> @ExperimentDecorator("my_experiment", seed=42)
        >> @ParamDecorator("my_param", default=1, description="My param description")
        >> def train(params):
        >>     pass

        that is equivalent to:
        >> def train(params):
        >>     pass
        >>
        >> train = ParamDecorator("my_param", default=1, description="My param description")(train)
        >> train = ExperimentDecorator("my_experiment")(train)

        Params of the experiment are accessible via train.params.


        """
        name = check_name(name)
        if name in _experiments:
            raise ValueError(f"Experiment with name {name} already exists")

        _experiments[name] = self

        self.name = name
        self.params = list()
        if seed is not None:
            self.params.append(Param.from_values(name="seed", default=seed, type=int))
        else:
            self.params.append(Param.from_values(name="seed", type=int, required=True, default=42))
        self.train = lambda x: print("this shouldn't have been called at all")

    def __call__(self, func: Callable[..., None]) -> Optional[Callable[..., None]] :
        if type(func) == InnerWrap:
            # check if seed is in params:
            if not any(param.name == "seed" for param in func.param_set):
                # add seed param
                self.params.extend(list(func.param_set))
            else:
                self.params = list(func.param_set)
            self.train = func.func

            return

        if callable(func) and len(func.__code__.co_varnames) >= 1:
            self.train = func

            return
        else:
            raise ValueError(
                f"Experiment decorator can only be applied to a function that accepts one argument, "
                + f"or on top of another param decorator. \n\tHave: \t{self.name} \n\tFunc: \t{self.train}"
            )

    @classmethod
    def from_ast(
        cls, ast_exps: List[Tuple[Expdec, Dict[str, Paramdec]]]
    ) -> Dict[str, "ExperimentDecorator"]:
        experiments = {}
        for ast_exp, ast_params in ast_exps:
            exp = cls(ast_exp.arg_pairs["name"])
            for ast_param in ast_params.values():
                param = ParamDecorator.from_ast(ast_param)
                exp.params.append(param.param)
            if exp.name in experiments:
                raise ValueError(f"Experiment with name {exp.name} already exists")
            experiments[exp.name] = exp

        return experiments


class ParamDecorator:
    param: Param

    def __init__(
        self,
        name: str,
        *,
        default: Any  = None,
        description: Optional[str] = None,
        required: bool = False,
        type: Optional[Type] = None,
        options: Optional[Union[Tuple[Any, ...], List[Any]]] = None,
    ):
        self.param = Param.from_values(
            name=name,
            default=default,
            description=description,
            required=required,
            type=type,
            options=tuple(options) if options is not None else None,
        )

    def __call__(self, func: Callable[..., None]) -> Callable:
        if type(func) == InnerWrap:
            func.add_param(self.param)
            return func

        if callable(func) and len(func.__code__.co_varnames) >= 1:
            params = set()
            params.add(self.param)
            return InnerWrap(func, params)  # type: ignore
        else:
            raise ValueError(
                "Param decorator can only be applied to a function that accepts one argument, or on top of another param decorator."
            )

    @classmethod
    def list_from_ast(cls, ast_params: Dict[str, Paramdec]):
        params = list()
        for ast_param in ast_params.values():
            param = ParamDecorator.from_ast(ast_param)
            params.append(param)
        return params

    @classmethod
    def from_ast(cls, ast_param: Paramdec):
        return cls(**ast_param.arg_pairs)


_experiments: Dict[str, ExperimentDecorator] = {}


experiment = ExperimentDecorator
param = ParamDecorator



from datasets import load_dataset 

class AlpacaDataset:
    def __init__(self, dataset_name, split="train"):
        self.dataset = load_dataset(dataset_name, split=split)
        
    def __len__(self):
        return len(self.dataset)
        
    def __getitem__(self, idx):
        item = self.dataset[idx]
        
        instruction = item["instruction"]
        
        if "input" in item.keys():
            prompt = (
                "Below is an instruction that describes a task. "
                "Write a response that appropriately completes the request.\n\n"
                f"### Instruction:\n{instruction}\n\n### Response:"
            )
        else:
            input = item["input"]
            
            prompt = (
                "Below is an instruction that describes a task, paired with an input that provides further context. "
                "Write a response that appropriately completes the request.\n\n"
                f"### Instruction:\n{instruction}\n\n### Input:\n{input}\n\n### Response:"
            )
            
        completion = item["output"]
        
        return {
            "prompt": prompt,
            "completion": completion,
        }


from typing import List, Optional
from pathlib import Path
from importlib.machinery import SourceFileLoader
import os
import dotenv

import subprocess


def get_key_from_path_or_key(key_or_path: Optional[str]) -> str:
    if key_or_path is None or key_or_path == "":
        raise ValueError("SSH_KEY in env is None")

    if (
        key_or_path.startswith("-----BEGIN")
        or key_or_path.startswith("ssh-rsa")
        or key_or_path.startswith("NOTHING")
    ):
        return key_or_path

    path = Path(os.path.expanduser(key_or_path)).resolve()
    if path.exists():
        return path.read_text()

    return key_or_path


class AppConfig:
    name: str
    github_repo_url: Optional[str] = None
    hosts: List[str]
    user: str
    key: str
    port: int
    number_of_processes_per_node: int

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)
    @classmethod
    def from_path(cls, path: Path) -> "AppConfig":
        config_path = path / "src" / "config.py"
        if not config_path.exists():
            raise ValueError(f"Config file {config_path} not found")

        if not (path / "env").exists():
            raise ValueError(f"Env file {path/ 'env'} not found")
        dotenv.load_dotenv(path / "env", verbose=True, override=True)

        try:
            module = SourceFileLoader("module.name", str(config_path)).load_module()
        except Exception as e:
            raise ValueError(
                f"Config file {config_path} cannot be loaded since your file doesn't meet requirements"
            ) from e

        name = str(module.__dict__["NAME"])
        github_repo_url: str = module.__dict__.get("GITHUB_REPO_URL", None)
        hosts = [host.strip() for host in module.__dict__["HOSTS"]]
        user = module.__dict__["HOSTS_USER"]

        if user == "root":
            raise ValueError("Please don't use root as the user")

        port = module.__dict__["HOSTS_PORT"]
        number_of_processes_per_node = module.__dict__["NUMBER_OF_PROCESSES_PER_NODE"]

        key = get_key_from_path_or_key(os.getenv("SSH_KEY"))

        return AppConfig(
            name=name,
            github_repo_url=github_repo_url,
            hosts=hosts,
            user=user,
            key=key,
            port=port,
            number_of_processes_per_node=number_of_processes_per_node,
        )

    def get_git_origin_url(self, path) -> Optional[str]:
        if self.github_repo_url is not None:
            return self.github_repo_url
        try:
            # Run the Git command to get the remote origin URL
            result = subprocess.check_output(
                ["git", "config", "--get", "remote.origin.url"],
                cwd=path,
                universal_newlines=True,
            )

            # Strip any leading/trailing whitespace from the result
            origin_url = result.strip()

            return origin_url
        except subprocess.CalledProcessError:
            return None

    def set_git_origin_url(self, path: Path):
        config_path = path / "src" / "config.py"
        # remove the GITHUB_REPO_URL line
        with open(config_path, "r") as f:
            lines = f.readlines()

        # Remove all occurrences of GITHUB_REPO_URL
        lines = [line for line in lines if "GITHUB_REPO_URL" not in line]

        with open(config_path, "w") as f:
            # Write back the modified lines
            f.writelines(lines)
    
            if lines[-1] != "\n":
                f.write("\n")

        # Add GITHUB_REPO_URL at the end without increasing new lines
        with open(config_path, "a") as f:
            f.write(f'GITHUB_REPO_URL = "{self.github_repo_url}"\n')

    def is_valid(self) -> Optional[str]:
        if self.github_repo_url is None:
            return "GITHUB_REPO_URL is None"
        if self.key is None:
            return "SSH_KEY in env is None"
        return



import ast

from typing import Any, List, Dict, Tuple, Optional


def func_defs(module: ast.Module) -> List[ast.FunctionDef]:
    defs = []
    for node in ast.iter_child_nodes(module):
        if isinstance(node, ast.FunctionDef):
            defs.append(node)

    return defs


def filter_experiment_defs(defs: List[ast.FunctionDef]):
    filtered_defs: List[ast.FunctionDef] = []
    for node in defs:
        try:
            if (
                len(node.decorator_list) >= 1
                and node.decorator_list[0].func.id == "experiment"  # type: ignore
            ):
                filtered_defs.append(node)
        except Exception:
            pass
    return filtered_defs


class Dec:
    name: str
    arg_pairs: Dict[str, Any]
    allowed_args = dict()

    def __init__(
        self,
        name: str,
        allowed_args: Dict[str, Tuple[type, ...]],
        arg_pairs: Optional[Dict[str, Any]] = None,
    ):
        self.name = name
        self.allowed_args = allowed_args
        self.arg_pairs = dict() if arg_pairs is None else arg_pairs

    def add_arg_pair(self, left: str, right: Any):
        if left not in self.allowed_args:
            raise ValueError(f"argument {left} of {self.name} is not allowed")
        if type(right) not in self.allowed_args[left]:
            raise ValueError(
                f"argument {left} of {self.name} has type {type(right)}, need {self.allowed_args[left]}"
            )

        if left in self.arg_pairs:
            raise ValueError(f"argument {left} is redefined in {self.name}")

        self.arg_pairs[left] = right


noneType = type(None)


class Expdec(Dec):
    def __init__(self):
        super(Expdec, self).__init__(
            name="experiment", allowed_args={"name": (str,), "seed": (int, noneType)}
        )


class Paramdec(Dec):
    def __init__(self):
        super(Paramdec, self).__init__(
            name="param",
            allowed_args={
                "name": (str,),
                "default": (str, int, bool, float, noneType),
                "description": (str, noneType),
                "required": (bool, noneType),
                "type": (type,),
                "options": (tuple, noneType),
            },
        )


builder = {"experiment": Expdec, "param": Paramdec}

type_dict = {
    "str": str,
    "int": int,
    "float": float,
    "bool": bool,
}


def build_experiment_def(
    node: ast.FunctionDef,
) -> Optional[Tuple[Expdec, Dict[str, Paramdec]]]:
    experiment: Optional[Expdec] = None
    params: Dict[str, Paramdec] = dict()
    stop = False

    for maybe_decorator in node.decorator_list:
        if not isinstance(maybe_decorator, ast.Call):
            continue
        decorator: ast.Call = maybe_decorator  # type: ignore
        func: Optional[ast.Name] = getattr(decorator, "func", None)
        if func is None:
            stop = True
            break
        dec_to_call = builder.get(func.id, None)

        if dec_to_call is None:
            stop = True
            break

        dec = dec_to_call()

        if len(decorator.args) > 1:
            raise ValueError(
                "experiment or param decorators cannot "
                + "have other decorators applied or unnamed params other than name"
            )

        if len(decorator.args) == 1:
            try:
                dec.add_arg_pair("name", decorator.args[0].value)  # type: ignore
            except Exception:
                stop = True
                break

        for kw in decorator.keywords:
            try:
                field_val = None
                if isinstance(kw.value, ast.Name):
                    field_val = type_dict[kw.value.id]
                elif isinstance(kw.value, ast.Constant):
                    field_val = kw.value.value
                elif isinstance(kw.value, ast.Tuple) or isinstance(kw.value, ast.List):
                    vals = []
                    for elt in kw.value.elts:
                        vals.append(elt.value)  # type: ignore
                    field_val = tuple(vals)

                else:
                    raise ValueError(
                        f"cannot find the type of kw.value: {type(kw.value)}"
                    )
                dec.add_arg_pair(kw.arg, field_val)  # type: ignore
            except Exception as e:
                print(e)
                stop = True
                break

        if type(dec) == Expdec:
            if experiment is not None:
                raise ValueError("more than one experiment is defined")
            else:
                experiment = dec

        if type(dec) == Paramdec:
            if dec.arg_pairs["name"] in params:
                raise ValueError(
                    f'more than one param with the same name {dec.arg_pairs["name"]} is defined'
                )
            params[dec.arg_pairs["name"]] = dec
    if stop:
        return

    if experiment is None:
        return

    return experiment, params


def build_experiment_defs(
    defs: List[ast.FunctionDef],
) -> List[Tuple[Expdec, Dict[str, Paramdec]]]:
    exps: List[Tuple[Expdec, Dict[str, Paramdec]]] = list()
    for node in defs:
        ret = build_experiment_def(node)
        if ret is not None:
            exp, params = ret
            exps.append((exp, params))

    return exps


def parse_experiments(filename: str) -> List[Tuple[Expdec, Dict[str, Paramdec]]]:
    parsed_code: Optional[ast.Module] = None
    with open(filename, "r") as f:
        parsed_code = ast.parse(f.read())

    if parsed_code is None:
        return []

    # search for top level experiment declarations
    defs = func_defs(parsed_code)

    # filter out by decorators, top level decorator should be "experiment"
    defs = filter_experiment_defs(defs)

    got = build_experiment_defs(defs)

    return got
